#ifndef DATABUFFER_H
#define DATABUFFER_H

#include <stdio.h>
#include <vector>
#include <stdexcept>
#include <string.h>
#include <string>

using std::runtime_error ;

#include <iostream>
using namespace std ;

#ifdef WIN32
#define snprintf _snprintf
#endif

struct DataBuffer
{
    enum Control {Reset} ;
    
    template <typename T>struct Binary
    {
        Binary(T t):data(t){}
        T data ;
    };
    
    struct Pointer
    {
        Pointer(const void *p, int s):ptr(p), sizeBytes(s) {}
        const void *ptr ;
        int sizeBytes ;
    };
    
    DataBuffer(int initialSizeBytes=64, int l=(1<<16)): byteoffset(0), totalbytes(initialSizeBytes), limit(l)
    {
        if(totalbytes == 0) throw std::runtime_error("Data buffer must be created with at least one byte!");
        
        buffer = new char [totalbytes] ;
        memset(buffer, 0, totalbytes);
    }
    
    DataBuffer(const char *data):limit(1<<16)
    {
        byteoffset = strlen(data) ;
        totalbytes = byteoffset + 1 ;
        if(byteoffset == 0) throw std::runtime_error("Data buffer must be created with at least one byte!");
        buffer = new char [totalbytes] ;
        strcpy(buffer, data);
    }
    
    DataBuffer(const char data):limit(1<<16)
    {
        byteoffset = 1 ;
        totalbytes = 2 ;
        buffer = new char [totalbytes] ;
        buffer [0] = data ;
    }
    
    DataBuffer(const DataBuffer &src):
        byteoffset(src.byteoffset),
        totalbytes(src.totalbytes),
        limit(src.limit)
    {
        buffer = new char [totalbytes] ;
        memcpy(buffer, src.buffer, totalbytes);
    }
    
    ~DataBuffer()
    {
        delete [] buffer ;
    }
    
    inline void reset() { byteoffset = 0x0 ; }
    inline char * operator * () const { return buffer ; }
    
    inline char & operator [] (unsigned i)
    {
        if(i >= totalbytes) throw runtime_error("buffer index exceeds range!");
        return buffer [i] ;
    }
    
    inline bool operator == (const DataBuffer &db)
    {
        if(byteoffset != db.byteoffset) return false ;
        for(int i=0; i<byteoffset; i++)
            if(buffer [i] != db.buffer [i])
                return false ;
        return true ;
    }
    
    inline DataBuffer & operator << (Control c)
    {
        if(c == DataBuffer::Reset)
            reset();
        else
            throw runtime_error("unrecognised control sequence!");
        return *this ;
    }
    
    inline DataBuffer & operator << (const char *data)
    {
        unsigned strSize = ::strlen(data) ;
        //Don't forget the terminating character!
        expansioncheck(strSize + 1);
        strcpy(&buffer [byteoffset], data) ;
        byteoffset += strSize ;
        return *this ;
    }
    
    inline DataBuffer & operator << (string &data)
    {
        unsigned strSize = data.size();
        //Don't forget the terminating character!
        expansioncheck(strSize + 1);
        strcpy(&buffer [byteoffset], data.c_str()) ;
        byteoffset += strSize ;
        return *this ;
    }
    
    inline DataBuffer & operator << (int data)
    {
        //ensure we've got plenty of room for the integer
        expansioncheck(16);
        int bytesUsed = 0x0 ;
        if((bytesUsed=snprintf(&buffer [byteoffset], (totalbytes-byteoffset), "%i", data)) < 0)
            throw runtime_error("Failed to format the integer");
        byteoffset += bytesUsed ;
        return *this ;
    }

    inline DataBuffer & operator << (double data)
    {
        //leave plenty of room for the double
        expansioncheck(64);
        int bytesUsed = 0x0 ;
        if((bytesUsed=snprintf(&buffer [byteoffset], (totalbytes-byteoffset), "%f", data)) < 0)
            throw runtime_error("Failed to format the double");
        byteoffset += bytesUsed ;
        return *this ;
    }
    
    inline DataBuffer & operator << (char data)
    {
        //leave plenty of room for the double
        expansioncheck(2);
        buffer [byteoffset] = data ;
        byteoffset += 1 ;
        return *this ;
    }
    
    inline DataBuffer & operator << (const DataBuffer &source)
    {
        //leave plenty of room for the double
        expansioncheck(source.byteoffset);
        memcpy(&buffer[byteoffset], *source, source.byteoffset);
        byteoffset += source.byteoffset ;
        return *this ;
    }
    
    inline DataBuffer & operator << (const DataBuffer::Pointer &source)
    {
        //leave plenty of room for the double
        expansioncheck(source.sizeBytes);
        memcpy(&buffer[byteoffset], source.ptr, source.sizeBytes);
        byteoffset += source.sizeBytes ;
        return *this ;
    }
    
    template <typename T> inline DataBuffer & operator << (const DataBuffer::Binary<T> &source)
    {
        appendBinary(source.data);
    }
    
    template <typename T> inline DataBuffer &  appendBinary(T t)
    {
        int sizeBytes = sizeof(T);
        expansioncheck(sizeBytes+1);
        memcpy(&buffer[byteoffset], &t, sizeBytes);
        byteoffset += sizeBytes ;
        return *this ;
    }
    
    inline  void expansioncheck(unsigned sizeBytes)
    {
        if(byteoffset + sizeBytes < totalbytes)
            return ;
        
        if(byteoffset + sizeBytes > limit)
        {
            throw std::runtime_error("Expansion has exceeded the allowed limit of the data buffer!");
        }
        
        int tmpSize = totalbytes*2 ;
        while(tmpSize < sizeBytes) tmpSize *= 2 ;
        if(tmpSize > limit)
        {
            tmpSize = limit ;
        }
        
        char *tmpBuffer = new char [tmpSize] ;
        memset(tmpBuffer, 0, tmpSize);
        memcpy(tmpBuffer, buffer, totalbytes);
        delete [] buffer ;
        buffer = tmpBuffer ;
        totalbytes = tmpSize ;
    }
    
    int byteoffset, totalbytes, limit ;
    char *buffer ;
};

#ifdef WIN32
#undef snprintf
#endif

#endif

