#ifndef HASH_MAP_H
#define HASH_MAP_H

#define uint32_t unsigned int
#define uint16_t unsigned short
#define uint8_t  unsigned char

#define hashsize(n) ((uint32_t)1<<(n))
#define hashmask(n) (hashsize(n)-1)
#define rot(x,k) (((x)<<(k)) | ((x)>>(32-(k))))

/*
Hashing functions taken from lookup3.c, by Bob Jenkins, May 2006, Public Domain.
*/

/*
 * My best guess at if you are big-endian or little-endian.  This may
 * need adjustment.
 */
 
#if (defined(__BYTE_ORDER) && defined(__LITTLE_ENDIAN) && \
     __BYTE_ORDER == __LITTLE_ENDIAN) || \
    (defined(i386) || defined(__i386__) || defined(__i486__) || \
     defined(__i586__) || defined(__i686__) || defined(vax) || defined(MIPSEL))
# define HASH_LITTLE_ENDIAN 1
# define HASH_BIG_ENDIAN 0
#elif (defined(__BYTE_ORDER) && defined(__BIG_ENDIAN) && \
       __BYTE_ORDER == __BIG_ENDIAN) || \
      (defined(sparc) || defined(POWERPC) || defined(mc68000) || defined(sel))
# define HASH_LITTLE_ENDIAN 0
# define HASH_BIG_ENDIAN 1
#else
# define HASH_LITTLE_ENDIAN 0
# define HASH_BIG_ENDIAN 0
#endif

#include <stdio.h>      /* defines printf for tests */
#include <time.h>       /* defines time_t for timings in the test */
//#include <stdint.h>     /* defines uint32_t etc */
//#include <sys/param.h>  /* attempt to define endianness */
#ifdef linux
# include <endian.h>    /* attempt to define endianness */
#endif

#define mix(a,b,c) \
{ \
  a -= c;  a ^= rot(c, 4);  c += b; \
  b -= a;  b ^= rot(a, 6);  a += c; \
  c -= b;  c ^= rot(b, 8);  b += a; \
  a -= c;  a ^= rot(c,16);  c += b; \
  b -= a;  b ^= rot(a,19);  a += c; \
  c -= b;  c ^= rot(b, 4);  b += a; \
}

#define final(a,b,c) \
{ \
  c ^= b; c -= rot(b,14); \
  a ^= c; a -= rot(c,11); \
  b ^= a; b -= rot(a,25); \
  c ^= b; c -= rot(b,16); \
  a ^= c; a -= rot(c,4);  \
  b ^= a; b -= rot(a,14); \
  c ^= b; c -= rot(b,24); \
}

static uint32_t hashword(
const uint32_t *k,                   /* the key, an array of uint32_t values */
size_t          length,               /* the length of the key, in uint32_ts */
uint32_t        initval)         /* the previous hash, or an arbitrary value */
{
  uint32_t a,b,c;

  /* Set up the internal state */
  a = b = c = 0xdeadbeef + (((uint32_t)length)<<2) + initval;

  /*------------------------------------------------- handle most of the key */
  while (length > 3)
  {
    a += k[0];
    b += k[1];
    c += k[2];
    mix(a,b,c);
    length -= 3;
    k += 3;
  }

  /*------------------------------------------- handle the last 3 uint32_t's */
  switch(length)                     /* all the case statements fall through */
  { 
  case 3 : c+=k[2];
  case 2 : b+=k[1];
  case 1 : a+=k[0];
    final(a,b,c);
  case 0:     /* case 0: nothing left to add */
    break;
  }
  /*------------------------------------------------------ report the result */
  return c;
}

static uint32_t hashlittle( const void *key, size_t length, uint32_t initval)
{
  uint32_t a,b,c;                                          /* internal state */
  union { const void *ptr; size_t i; } u;     /* needed for Mac Powerbook G4 */

  /* Set up the internal state */
  a = b = c = 0xdeadbeef + ((uint32_t)length) + initval;

  u.ptr = key;
  if (HASH_LITTLE_ENDIAN && ((u.i & 0x3) == 0)) {
    const uint32_t *k = (const uint32_t *)key;         /* read 32-bit chunks */
    #ifdef VALGRIND
    const uint8_t  *k8;
    #endif

    /*------ all but last block: aligned reads and affect 32 bits of (a,b,c) */
    while (length > 12)
    {
      a += k[0];
      b += k[1];
      c += k[2];
      mix(a,b,c);
      length -= 12;
      k += 3;
    }

    /*----------------------------- handle the last (probably partial) block */
    /* 
     * "k[2]&0xffffff" actually reads beyond the end of the string, but
     * then masks off the part it's not allowed to read.  Because the
     * string is aligned, the masked-off tail is in the same word as the
     * rest of the string.  Every machine with memory protection I've seen
     * does it on word boundaries, so is OK with this.  But VALGRIND will
     * still catch it and complain.  The masking trick does make the hash
     * noticably faster for short strings (like English words).
     */
#ifndef VALGRIND

    switch(length)
    {
    case 12: c+=k[2]; b+=k[1]; a+=k[0]; break;
    case 11: c+=k[2]&0xffffff; b+=k[1]; a+=k[0]; break;
    case 10: c+=k[2]&0xffff; b+=k[1]; a+=k[0]; break;
    case 9 : c+=k[2]&0xff; b+=k[1]; a+=k[0]; break;
    case 8 : b+=k[1]; a+=k[0]; break;
    case 7 : b+=k[1]&0xffffff; a+=k[0]; break;
    case 6 : b+=k[1]&0xffff; a+=k[0]; break;
    case 5 : b+=k[1]&0xff; a+=k[0]; break;
    case 4 : a+=k[0]; break;
    case 3 : a+=k[0]&0xffffff; break;
    case 2 : a+=k[0]&0xffff; break;
    case 1 : a+=k[0]&0xff; break;
    case 0 : return c;              /* zero length strings require no mixing */
    }

#else /* make valgrind happy */

    k8 = (const uint8_t *)k;
    switch(length)
    {
    case 12: c+=k[2]; b+=k[1]; a+=k[0]; break;
    case 11: c+=((uint32_t)k8[10])<<16;  /* fall through */
    case 10: c+=((uint32_t)k8[9])<<8;    /* fall through */
    case 9 : c+=k8[8];                   /* fall through */
    case 8 : b+=k[1]; a+=k[0]; break;
    case 7 : b+=((uint32_t)k8[6])<<16;   /* fall through */
    case 6 : b+=((uint32_t)k8[5])<<8;    /* fall through */
    case 5 : b+=k8[4];                   /* fall through */
    case 4 : a+=k[0]; break;
    case 3 : a+=((uint32_t)k8[2])<<16;   /* fall through */
    case 2 : a+=((uint32_t)k8[1])<<8;    /* fall through */
    case 1 : a+=k8[0]; break;
    case 0 : return c;
    }

#endif /* !valgrind */

  } else if (HASH_LITTLE_ENDIAN && ((u.i & 0x1) == 0)) {
    const uint16_t *k = (const uint16_t *)key;         /* read 16-bit chunks */
    const uint8_t  *k8;

    /*--------------- all but last block: aligned reads and different mixing */
    while (length > 12)
    {
      a += k[0] + (((uint32_t)k[1])<<16);
      b += k[2] + (((uint32_t)k[3])<<16);
      c += k[4] + (((uint32_t)k[5])<<16);
      mix(a,b,c);
      length -= 12;
      k += 6;
    }

    /*----------------------------- handle the last (probably partial) block */
    k8 = (const uint8_t *)k;
    switch(length)
    {
    case 12: c+=k[4]+(((uint32_t)k[5])<<16);
             b+=k[2]+(((uint32_t)k[3])<<16);
             a+=k[0]+(((uint32_t)k[1])<<16);
             break;
    case 11: c+=((uint32_t)k8[10])<<16;     /* fall through */
    case 10: c+=k[4];
             b+=k[2]+(((uint32_t)k[3])<<16);
             a+=k[0]+(((uint32_t)k[1])<<16);
             break;
    case 9 : c+=k8[8];                      /* fall through */
    case 8 : b+=k[2]+(((uint32_t)k[3])<<16);
             a+=k[0]+(((uint32_t)k[1])<<16);
             break;
    case 7 : b+=((uint32_t)k8[6])<<16;      /* fall through */
    case 6 : b+=k[2];
             a+=k[0]+(((uint32_t)k[1])<<16);
             break;
    case 5 : b+=k8[4];                      /* fall through */
    case 4 : a+=k[0]+(((uint32_t)k[1])<<16);
             break;
    case 3 : a+=((uint32_t)k8[2])<<16;      /* fall through */
    case 2 : a+=k[0];
             break;
    case 1 : a+=k8[0];
             break;
    case 0 : return c;                     /* zero length requires no mixing */
    }

  } else {                        /* need to read the key one byte at a time */
    const uint8_t *k = (const uint8_t *)key;

    /*--------------- all but the last block: affect some 32 bits of (a,b,c) */
    while (length > 12)
    {
      a += k[0];
      a += ((uint32_t)k[1])<<8;
      a += ((uint32_t)k[2])<<16;
      a += ((uint32_t)k[3])<<24;
      b += k[4];
      b += ((uint32_t)k[5])<<8;
      b += ((uint32_t)k[6])<<16;
      b += ((uint32_t)k[7])<<24;
      c += k[8];
      c += ((uint32_t)k[9])<<8;
      c += ((uint32_t)k[10])<<16;
      c += ((uint32_t)k[11])<<24;
      mix(a,b,c);
      length -= 12;
      k += 12;
    }

    /*-------------------------------- last block: affect all 32 bits of (c) */
    switch(length)                   /* all the case statements fall through */
    {
    case 12: c+=((uint32_t)k[11])<<24;
    case 11: c+=((uint32_t)k[10])<<16;
    case 10: c+=((uint32_t)k[9])<<8;
    case 9 : c+=k[8];
    case 8 : b+=((uint32_t)k[7])<<24;
    case 7 : b+=((uint32_t)k[6])<<16;
    case 6 : b+=((uint32_t)k[5])<<8;
    case 5 : b+=k[4];
    case 4 : a+=((uint32_t)k[3])<<24;
    case 3 : a+=((uint32_t)k[2])<<16;
    case 2 : a+=((uint32_t)k[1])<<8;
    case 1 : a+=k[0];
             break;
    case 0 : return c;
    }
  }

  final(a,b,c);
  return c;
}

#include <vector>
#include <list>
#include <assert.h>
#include <string>
#include <string.h>

using std::vector ;
using std::list ;
using std::string ;

static bool keyGreaterOverride(int a, int b) { return (a>b); }
static uint32_t keyHashOverride(int v) { return hashword((uint32_t *)&v, 1, 0); }
static uint32_t keyHashOverride(void *v) { return hashword((uint32_t *)&v, sizeof(void *)/4, 0); }
static uint32_t keyHashOverride(const char *s) { return hashlittle(s, strlen(s), 0); }
static uint32_t keyHashOverride(string v) { return hashlittle(v.c_str(), v.size(), 0); }
//uint32_t keyHashOverride(const string v) { return hashlittle(v.c_str(), v.size(), 0); }

#ifdef DEBUG
#include <iostream>
using namespace std ;
#endif 

template <typename KEY, typename VALUE> class HashMap
{
    public:
        
        class HashNode
        {
            public:
                HashNode(KEY k, VALUE v):key(k), value(v) {prev = next = NULL ;}
                KEY key ;
                VALUE value ;
                HashNode *prev ;
                HashNode *next ;
                
        };
        
        HashMap(int startSize=64, int maxOcc=4):_count(0), _maxOcc(maxOcc)
        {
            calcHashMask(startSize);
            first = last = iter = NULL ;
            _table = new vector < list <HashNode> > (startSize) ;
        }
        
        HashMap(const HashMap &src):_count(0), _maxOcc(src._maxOcc)
        {
            int startSize = src._table -> size() ;
            calcHashMask(startSize);
            first = last = iter = NULL ;
            _table = new vector < list <HashNode> > (startSize) ;
            
            HashNode *node = src.first ;
            while(node)
            {
                insert(node -> key, node -> value) ;
                node = node -> next ;
            }
        }
        
        ~HashMap()
        {
            #ifdef DEBUG
                cout<<"Deleting the map table ["<<_table<<"]...."<<endl;
            #endif 
            delete _table ;
        }
        
        HashMap & operator= (const HashMap &src)
        {
            if(_table -> size() != src._table -> size()) {
                delete _table ;
                int startSize = src._table -> size() ;
                calcHashMask(startSize);
                first = last = iter = NULL ;
                _table = new vector < list <HashNode> > (startSize) ;
            }
            else
            {
                clear();
            }
            _count = 0 ;
            HashNode *node = src.first ;
            while(node)
            {
                insert(node -> key, node -> value) ;
                node = node -> next ;
            }
        }
        
        void calcHashMask(int tableSize)
        {
            #ifdef DEBUG
                int origTableSize = tableSize ;
            #endif
            _hashMask = 0 ;
            int count = 0 ;
            for(int i=0; i<32; i++)
            {
                int res = (tableSize & 1) ;
                if(res)
                {
                    _hashMask = hashmask(i) ;
                }
                count += res ;
                tableSize >>= 1 ;
            }
            #ifdef DEBUG
                cout<<"calcHashMask count: "<<count<<" for tableSize: "<<origTableSize<<endl;
            #endif 

            assert(count == 1 && "HashMap only supports table sizes that are powers of 2!");
        }
        
        
        inline bool contains(const KEY key) const { return hasKey(key); }
        
        bool hasKey(const KEY key) const
        {
            if(_count == 0) return false ;
            uint32_t hashValue = (keyHashOverride(key) & _hashMask) ;
            typename list <HashNode>::iterator x ;
            for(x=(*_table) [hashValue].begin(); x!=(*_table) [hashValue].end(); x++)
            {
                if(x -> key == key)
                    return true ;
            }
            return false ;
        }

        VALUE & operator [](const KEY key) const
        {
            uint32_t hashValue = (keyHashOverride(key) & _hashMask) ;
            typename list <HashNode>::iterator x ;
            for(x=(*_table) [hashValue].begin(); x!=(*_table) [hashValue].end(); x++)
            {
                if(x -> key == key)
                {
                    return x -> value ;
                }
            }
            assert(false && "HashMap [] operator called with a non-existant key!");
        }

        HashNode * get(const KEY key)
        {
            uint32_t hashValue = (keyHashOverride(key) & _hashMask) ;
            typename list <HashNode>::iterator x ;
            for(x=(*_table) [hashValue].begin(); x!=(*_table) [hashValue].end(); x++)
            {
                if(x -> key == key)
                {
                    return &(*x) ;
                }
            }
            assert(false && "HashMap get method called with a non-existant key!");
        }
        
        HashNode * next(const KEY key)
        {
            uint32_t hashValue = (keyHashOverride(key) & _hashMask) ;
            typename list <HashNode>::iterator x ;
            for(x=(*_table) [hashValue].begin(); x!=(*_table) [hashValue].end(); x++)
            {
                if(x -> key == key)
                {
                    return x -> next ;
                }
            }
            assert(false && "HashMap next method called with a non-existant key!");
        }
        
        //preserves insertion order
        void insert(const KEY key, const VALUE value)
        {
            uint32_t hashValue = (keyHashOverride(key) & _hashMask) ;
            typename list <HashNode>::iterator x ;
            for(x= (*_table) [hashValue].begin(); x!=(*_table) [hashValue].end(); x++)
            {
                if(x -> key == key)
                {
                    x -> value = value ;
                    return ;
                }
            }
            (*_table) [hashValue].push_back(HashNode(key, value));
            HashNode *node = &((*_table) [hashValue].back());
            
            if(last != NULL)
            {
                node -> prev = last ;
                last -> next = node ;
            }
            last = node ;
            if(first == NULL)
                first = node ;
            
            _count += 1 ;
            
            if((*_table) [hashValue].size() > _maxOcc)
            {
                resize();
            }
        }
        //preserves key order
        void insertAscending(const KEY key, const VALUE value)
        {
            uint32_t hashValue = (keyHashOverride(key) & _hashMask) ;
            typename list <HashNode>::iterator x ;
            for(x= (*_table) [hashValue].begin(); x!=(*_table) [hashValue].end(); x++)
            {
                if(x -> key == key)
                {
                    x -> value = value ;
                    return ;
                }
            }
            
            (*_table) [hashValue].push_back(HashNode(key, value));
            HashNode *node = &((*_table) [hashValue].back());
            bool assignedOK = false ;
            HashNode *iter = first ;
            
            while(iter)
            {
                if(keyGreaterOverride(iter -> key, node -> key))
                {
                    node -> next = iter ;
                    node -> prev = iter -> prev ;
                    if(node -> prev)
                        node -> prev -> next = node ;
                    else
                        first = node ;
                    iter -> prev = node ;
                    assignedOK = true ;
                    break ;
                }
                iter = iter -> next ;
            }
            
            if(assignedOK == false)
            {
                if(last != NULL)
                {
                    last -> next = node ;
                    node -> prev = last ;
                }
                last = node ;
            }
            
            if(first == NULL)
                first = node ;
            
            _count += 1 ;
            
            if((*_table) [hashValue].size() > _maxOcc)
            {
                resize();
            }
        }
        
        void remove(const KEY key)
        {
            uint32_t hashValue = (keyHashOverride(key) & _hashMask) ;
            typename list <HashNode>::iterator x ;
            for(x=(*_table) [hashValue].begin(); x!=(*_table) [hashValue].end(); x++)
            {
                if(x -> key == key)
                {
                    HashNode *node = &(*x) ;
                    //Rewire the pointers to maintain order
                    if(node -> prev == NULL) //first
                        first = node -> next ;
                    else
                        node -> prev -> next = node -> next ;
                    if(node -> next == NULL) // last
                        last = node -> prev ;
                    else
                        node -> next -> prev = node -> prev ;
                    _count -= 1 ;
                    (*_table) [hashValue].erase(x) ;
                    return ;
                }
            }
            assert(false && "HashMap::remove called with a non-existant key");
        }
        
        void clear()
        {
            _count = 0 ;
            for(int i=0; i<(*_table).size(); i++)
                (*_table) [i].clear();
            first = last = iter = NULL ;
        }
        
        void resize(uint32_t tableTargetSize=0)
        {
            if(tableTargetSize == 0)
                tableTargetSize = _table -> size() * 2 ;
            if(tableTargetSize == _table -> size())
                return ;
            
            assert((tableTargetSize > _table -> size()) && "Cannot resize the HashMap to a smaller size!");
            //Fairly inefficient but easy to implement :)
            // Save enough data to restore the original map
            HashNode *node = first ;
            vector < list <HashNode> > *oldTable = _table ;

            // Reset the map with the doubled dimensions
            int tableSize = tableTargetSize ;
            _table = new vector < list <HashNode> > (tableSize) ;
            first = last = iter = NULL ;
            calcHashMask(tableSize) ;
            
            #ifdef DEBUG
                cout<<"Growing the map ["<<tableSize<<"]...." ;
            #endif 
            _count = 0 ;
            //Restore the original data in the correct order using the new map
            while(node)
            {
                insert(node -> key, node -> value) ;
                node = node -> next ;
            }
            delete oldTable ;
            #ifdef DEBUG
                cout<<"done"<<endl;
            #endif
        }
        
        inline int size() const{ return _count ;}
        inline int capacity() const{ return _table->size() ;}
        inline int occupancy() const{ return _maxOcc ;}
        
        inline HashNode * currIter()
        {
            return iter ;
        }
        
        inline VALUE & curr()
        {
            return iter -> value;
        }
        
        inline HashNode * next()
        {
            if(!iter) 
                iter = first ;
            else
                iter = iter -> next ;
            return iter;
        }
        
        inline HashNode * prev()
        {
            if(!iter) 
                iter = last ;
            else
                iter = iter -> prev ;
            return iter;
        }
        
        inline void resetIter()
        {
            iter = NULL ;
        }
        
        HashNode *first, *last, *iter ;
        uint32_t _count, _maxOcc, _hashMask ;
        vector < list <HashNode> > *_table ;
};

#undef uint32_t
#undef uint16_t
#undef uint8_t

#endif

