#include <libssh/libssh.h>
#include <boost/asio.hpp>
#include <boost/bind.hpp>
#include <DataBuffer.h>
#include <stdexcept>
#include <HashMap.h>
#include <iostream>
#include <assert.h>
#include <string>


using namespace std ;
using namespace boost::asio ;
using namespace boost::asio::ip ;

unsigned int DummyBuffer ; //

struct SshBoostDriver ;

typedef const boost::system::error_code SystemError ;

struct SshBoostDriver
{
    
    struct ChannelSession
    {
        ChannelSession(boost::asio::io_service &ios, ssh_channel c):socket(ios), channel(c), socketBuffer(8192), channelBuffer(8192) {}
        ~ChannelSession()
        {
            try
            {
                if(socket.is_open())
                {
                    socket.close();
                }
            }
            catch (boost::exception& e)
            {
                //we don't care in the destructor, just note it and carry on regardless
                cout<<"ChannelSession generated an exception when shutting down the socket!"<<endl;
            }
            if(ssh_channel_is_open(channel))
            {
                ssh_channel_close(channel);
            }
            ssh_channel_free(channel);
        }
        
        char * operator * () { return *socketBuffer; }
        int buffersize() { return socketBuffer.totalbytes; }
        
        tcp::socket socket ;
        ssh_channel channel ;
        DataBuffer socketBuffer, channelBuffer ;
    };
    
    SshBoostDriver(const char *hostname, int portNumber, const char *target, const char * tgtPort, boost::asio::io_service &ios):
        session(0x0),
        targetHost(target),
        targetPort(tgtPort),
        ioService(ios)
    {
        session = ssh_new();
        
        ssh_options_set(session, SSH_OPTIONS_HOST, hostname);
        ssh_options_set(session, SSH_OPTIONS_PORT, &portNumber);
        //ssh_options_set(session, SSH_OPTIONS_LOG_VERBOSITY_STR, "3");
    }
    
    void connectAuthPasswd(const char *username, const char *passwd)
    {
        ssh_options_set(session, SSH_OPTIONS_USER, username);
        if(ssh_connect(session) == SSH_ERROR)
        {
            throw std::runtime_error("Failed to connect to the host");
        }
        if(ssh_userauth_password(session,NULL,passwd) != SSH_AUTH_SUCCESS)
        {
            throw std::runtime_error("Failed to authenticate using username/password pair");
        }
        boost::asio::ip::tcp protocol(boost::asio::ip::tcp::v4()) ; // the infamous protocol parameter, at last.
        sessionSocket = boost::shared_ptr<tcp::socket>(new tcp::socket(ioService, protocol, ssh_get_fd(session)));
    }
    
    void registerReversePort(const char *address, int port)
    {
        int boundport(0x0);
        if(ssh_forward_listen(session, address, port, &boundport) == SSH_ERROR)
        {
            throw std::runtime_error("Failed to register the remote listener");
        }
    }
    
    void registerAsyncHandler()
    {
        if(!sessionSocket.get()) throw std::runtime_error("ssh connection has not been established!");
        int flags(boost::asio::detail::message_peek) ;
        sessionSocket->async_receive(boost::asio::buffer(&DummyBuffer, 1), flags,
                                    boost::bind(&SshBoostDriver::handleSession,
                                                this, boost::asio::placeholders::error, boost::asio::placeholders::bytes_transferred));
    }
    
    void handleSession(const SystemError& error, std::size_t bytesTransferred)
    {
        if(error || !bytesTransferred) throw std::runtime_error("ssh connection error!");
        
        ssh_channel tgtChannel(0x0) ;
        if((tgtChannel = ssh_forward_accept(session, 0)))
        {
            connectToTarget(tgtChannel);
        }
        else
        {
            vector <boost::shared_ptr<ChannelSession> > deadChannels ;
            for(sessionMap.resetIter(); sessionMap.next();)
            {
                ssh_channel channel = sessionMap.currIter()->key ;
                boost::shared_ptr<ChannelSession> &channelSession = sessionMap.currIter()->value ;
                DataBuffer &channelBuffer = channelSession->channelBuffer ;
                tcp::socket &socket       = channelSession->socket ;
                
                bool channelDead = (ssh_channel_is_eof(channel) || ssh_channel_is_closed(channel)) ;
                
                //We have to read all the data from the channel in this handler
                //as we only get back here when there's activity on the socket
                int bytesRead;
                while(!channelDead && (bytesRead = ssh_channel_read_nonblocking(channel, *channelBuffer, channelBuffer.totalbytes, 0)))
                {
                    //We send the data on immediately using a blocking syncrhonous call
                    try { socket.send(boost::asio::buffer(*channelBuffer, bytesRead)); }
                    catch (boost::exception& e) { channelDead = true; }
                }
                
                if(channelDead)
                {
                    deadChannels.push_back(channelSession);
                }
            }
            
            for(int ci=0; ci<deadChannels.size(); ci++)
            {
                closeChannel(deadChannels [ci]) ;
            }
        }
        registerAsyncHandler();
    }
    
    void connectToTarget(ssh_channel tgtChannel)
    {
        //connect
        ip::tcp::resolver resolver(ioService);
        ip::tcp::resolver::query query(targetHost, targetPort);
        ip::tcp::resolver::iterator iterator = resolver.resolve(query);
        ip::tcp::endpoint endpoint(*iterator);
        
        boost::shared_ptr<ChannelSession> channelSession(new ChannelSession(ioService, tgtChannel));
        channelSession->socket.connect(endpoint);
        sessionMap.insert(tgtChannel, channelSession);
        registerSocketAsync(channelSession);
    }
    
    void registerSocketAsync(boost::shared_ptr<ChannelSession> channelSession)
    {
        channelSession->socket.async_receive(boost::asio::buffer(**channelSession, channelSession->buffersize()),
                                             boost::bind(&SshBoostDriver::handleSocketRecv,
                                             this, boost::asio::placeholders::error, boost::asio::placeholders::bytes_transferred, channelSession));
    }
    
    void handleSocketRecv(const SystemError &error, size_t bytesTransferred, boost::shared_ptr<ChannelSession> channelSession)
    {
        if(error || !bytesTransferred)
        {
            //shutdown the socket and associated channel (if it's still open)
            closeChannel(channelSession);
        }
        else
        {
            if(ssh_channel_write(channelSession->channel, **channelSession, bytesTransferred) == SSH_ERROR)
            {
                closeChannel(channelSession);
            }
            else
            {
                registerSocketAsync(channelSession);
            }
        }
    }
    
    void closeChannel(boost::shared_ptr<ChannelSession> &channelSession)
    {
        ssh_channel channel = channelSession->channel ;
        
        if(!sessionMap.contains(channel)) return ; //channel has already been removed
        if(!ssh_channel_is_closed(channel)) ssh_channel_close(channel);
        
        tcp::socket &socket = channelSession->socket ;
        if(socket.is_open())
        {
            try { socket.close(); }
            catch (boost::exception& e) {} //we can't do anything for this exception
        }
        sessionMap.remove(channel);
    }
    
    string targetHost, targetPort ;
    boost::asio::io_service &ioService ;
    ssh_session session ;
    boost::shared_ptr<tcp::socket> sessionSocket ;
    HashMap <ssh_channel, boost::shared_ptr<ChannelSession> > sessionMap ;
};

int main()
{
    boost::asio::io_service ioService;
    SshBoostDriver sshDriver("gateway", 22, "service", "8080", ioService);
    
    sshDriver.connectAuthPasswd("username", "password");
    cout<<"Connected..."<<endl;
    sshDriver.registerReversePort("127.0.0.1", 8080);
    cout<<"Registered the remote listener..."<<endl;
    
    sshDriver.registerAsyncHandler();
    
    ioService.run();
    
    return 0 ;
}

